from xcollect.boutdataset import open_boutdataset


ds = open_boutdataset(datapath=".")

# To access just the desired numpy array values directly, slice then use .values
n = ds["n"].isel(t=slice(10, None, 5)).values

# The dataset can also be loaded without using dask, though this will be far more memory-intensive
ds = open_boutdataset(datapath=".", chunks=None)



import pytest

from pathlib import Path

import numpy.testing as npt
import xarray.testing as xrt

from xbout.tests.test_load import bout_xyt_example_files
from xbout import open_boutdataset


params_guards = "guards"
params_guards_values = [
    pytest.param({"x": 0, "y": 0}, marks=pytest.mark.long),
    pytest.param({"x": 2, "y": 0}, marks=pytest.mark.long),
    pytest.param({"x": 0, "y": 2}, marks=pytest.mark.long),
    {"x": 2, "y": 2},
]
params_boundaries = "keep_xboundaries, keep_yboundaries"
params_boundaries_values = [
    pytest.param(False, False, marks=pytest.mark.long),
    pytest.param(True, False, marks=pytest.mark.long),
    pytest.param(False, True, marks=pytest.mark.long),
    (True, True),
]


class TestRegion:
    @pytest.mark.long
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    def test_region_core(
        self, bout_xyt_example_files, guards, keep_xboundaries, keep_yboundaries
    ):
        # Note need to use more than (3*MXG,3*MYG) points per output file
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=4,
            nt=1,
            guards=guards,
            grid="grid",
            topology="core",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        n = ds["n"]

        n_core = n.bout.from_region("core")

        # Remove attributes that are expected to be different
        del n_core.attrs["regions"]
        del n.attrs["regions"]

        # Select only non-boundary data
        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        xrt.assert_identical(
            n.isel(theta=slice(ybndry, -ybndry if ybndry != 0 else None)),
            n_core.isel(
                theta=slice(guards["y"], -guards["y"] if guards["y"] != 0 else None)
            ),
        )

    @pytest.mark.long
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    def test_region_sol(
        self, bout_xyt_example_files, guards, keep_xboundaries, keep_yboundaries
    ):
        # Note need to use more than (3*MXG,3*MYG) points per output file
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=4,
            nt=1,
            guards=guards,
            grid="grid",
            topology="sol",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        n = ds["n"]

        n_sol = n.bout.from_region("SOL")

        # Remove attributes that are expected to be different
        del n_sol.attrs["regions"]
        del n.attrs["regions"]

        xrt.assert_identical(n, n_sol)

    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize("region_guards", [0, 1])
    def test_region_limiter(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        region_guards,
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=4,
            nt=1,
            guards=guards,
            grid="grid",
            topology="limiter",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]

        if keep_xboundaries:
            ixs = ds.metadata["ixseps1"]
        else:
            ixs = ds.metadata["ixseps1"] - guards["x"]

        # For selecting only non-boundary data
        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0

        n = ds["n"]
        n_noregions = n.copy(deep=False)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_sol = n.bout.from_region("SOL", with_guards=region_guards)

        # Remove attributes that are expected to be different
        # Corners may be different because core region 'communicates' in y
        del n_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs, None)),
            n_sol.isel(x=slice(region_guards, None)),
        )
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs - region_guards, ixs),
                theta=slice(ybndry, -ybndry if ybndry != 0 else None),
            ),
            n_sol.isel(
                x=slice(region_guards),
                theta=slice(ybndry, -ybndry if ybndry != 0 else None),
            ),
        )

        n_core = n.bout.from_region("core", with_guards=region_guards)

        # Remove attributes that are expected to be different
        del n_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs + region_guards),
                theta=slice(ybndry, -ybndry if ybndry != 0 else None),
            ),
            n_core.isel(
                theta=slice(
                    region_guards, -region_guards if region_guards != 0 else None
                )
            ),
        )

    @pytest.mark.long
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    def test_region_xpoint(
        self, bout_xyt_example_files, guards, keep_xboundaries, keep_yboundaries
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=4,
            nt=1,
            guards=guards,
            grid="grid",
            topology="xpoint",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs = ds.metadata["ixseps1"]
        else:
            ixs = ds.metadata["ixseps1"] - guards["x"]

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys1 = ds.metadata["jyseps1_1"] + ybndry
        ny_inner = ds.metadata["ny_inner"] + 2 * ybndry
        jys2 = ds.metadata["jyseps2_2"] + 3 * ybndry
        ny = ds.metadata["ny"] + 4 * ybndry

        n = ds["n"]
        n_noregions = n.copy(deep=True)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_lower_inner_PFR = n.bout.from_region("lower_inner_PFR")

        # Remove attributes that are expected to be different
        del n_lower_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys1 + 1)),
            n_lower_inner_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys2 + 1, jys2 + 1 + myg)
                ).values,
                n_lower_inner_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_lower_inner_SOL = n.bout.from_region("lower_inner_SOL")

        # Remove attributes that are expected to be different
        del n_lower_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys1 + 1)),
            n_lower_inner_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys1 + 1, jys1 + 1 + myg)
                ).values,
                n_lower_inner_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_upper_inner_PFR = n.bout.from_region("upper_inner_PFR")

        # Remove attributes that are expected to be different
        del n_upper_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys1 + 1, ny_inner)),
            n_upper_inner_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys2 + 1 - myg, jys2 + 1)
                ).values,
                n_upper_inner_PFR.isel(theta=slice(myg)).values,
            )

        n_upper_inner_SOL = n.bout.from_region("upper_inner_SOL")

        # Remove attributes that are expected to be different
        del n_upper_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys1 + 1, ny_inner)),
            n_upper_inner_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys1 + 1 - myg, jys1 + 1)
                ).values,
                n_upper_inner_SOL.isel(theta=slice(myg)).values,
            )

        n_upper_outer_PFR = n.bout.from_region("upper_outer_PFR")

        # Remove attributes that are expected to be different
        del n_upper_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(ny_inner, jys2 + 1)),
            n_upper_outer_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys1 + 1, jys1 + 1 + myg)
                ).values,
                n_upper_outer_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_upper_outer_SOL = n.bout.from_region("upper_outer_SOL")

        # Remove attributes that are expected to be different
        del n_upper_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(ny_inner, jys2 + 1)),
            n_upper_outer_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys2 + 1, jys2 + 1 + myg)
                ).values,
                n_upper_outer_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_lower_outer_PFR = n.bout.from_region("lower_outer_PFR")

        # Remove attributes that are expected to be different
        del n_lower_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys2 + 1, None)),
            n_lower_outer_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys1 + 1 - myg, jys1 + 1)
                ).values,
                n_lower_outer_PFR.isel(theta=slice(myg)).values,
            )

        n_lower_outer_SOL = n.bout.from_region("lower_outer_SOL")

        # Remove attributes that are expected to be different
        del n_lower_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys2 + 1, None)),
            n_lower_outer_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys2 + 1 - myg, jys2 + 1)
                ).values,
                n_lower_outer_SOL.isel(theta=slice(myg)).values,
            )

    @pytest.mark.long
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    def test_region_singlenull(
        self, bout_xyt_example_files, guards, keep_xboundaries, keep_yboundaries
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=4,
            nt=1,
            guards=guards,
            grid="grid",
            topology="single-null",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs = ds.metadata["ixseps1"]
        else:
            ixs = ds.metadata["ixseps1"] - guards["x"]

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys1 = ds.metadata["jyseps1_1"] + ybndry
        jys2 = ds.metadata["jyseps2_2"] + ybndry
        ny = ds.metadata["ny"] + 2 * ybndry

        n = ds["n"]
        n_noregions = n.copy(deep=True)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_inner_PFR = n.bout.from_region("inner_PFR")

        # Remove attributes that are expected to be different
        del n_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys1 + 1)),
            n_inner_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys2 + 1, jys2 + 1 + myg)
                ).values,
                n_inner_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_inner_SOL = n.bout.from_region("inner_SOL")

        # Remove attributes that are expected to be different
        del n_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys1 + 1)),
            n_inner_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys1 + 1, jys1 + 1 + myg)
                ).values,
                n_inner_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_core = n.bout.from_region("core")

        # Remove attributes that are expected to be different
        del n_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys1 + 1, jys2 + 1)),
            n_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys2 + 1 - myg, jys2 + 1)
                ).values,
                n_core.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys1 + 1, jys1 + 1 + myg)
                ).values,
                n_core.isel(theta=slice(-myg, None)).values,
            )

        n_sol = n.bout.from_region("SOL")

        # Remove attributes that are expected to be different
        del n_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys1 + 1, jys2 + 1)),
            n_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys1 + 1 - myg, jys1 + 1)
                ).values,
                n_sol.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys2 + 1, jys2 + 1 + myg)
                ).values,
                n_sol.isel(theta=slice(-myg, None)).values,
            )

        n_outer_PFR = n.bout.from_region("outer_PFR")

        # Remove attributes that are expected to be different
        del n_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys2 + 1, None)),
            n_outer_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys1 + 1 - myg, jys1 + 1)
                ).values,
                n_outer_PFR.isel(theta=slice(myg)).values,
            )

        n_outer_SOL = n.bout.from_region("outer_SOL")

        # Remove attributes that are expected to be different
        del n_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys2 + 1, None)),
            n_outer_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys2 + 1 - myg, jys2 + 1)
                ).values,
                n_outer_SOL.isel(theta=slice(myg)).values,
            )

    @pytest.mark.long
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    def test_region_connecteddoublenull(
        self, bout_xyt_example_files, guards, keep_xboundaries, keep_yboundaries
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards=guards,
            grid="grid",
            topology="connected-double-null",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs = ds.metadata["ixseps1"]
        else:
            ixs = ds.metadata["ixseps1"] - guards["x"]

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys11 = ds.metadata["jyseps1_1"] + ybndry
        jys21 = ds.metadata["jyseps2_1"] + ybndry
        ny_inner = ds.metadata["ny_inner"] + 2 * ybndry
        jys12 = ds.metadata["jyseps1_2"] + 3 * ybndry
        jys22 = ds.metadata["jyseps2_2"] + 3 * ybndry
        ny = ds.metadata["ny"] + 4 * ybndry

        n = ds["n"]
        n_noregions = n.copy(deep=True)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_lower_inner_PFR = n.bout.from_region("lower_inner_PFR")

        # Remove attributes that are expected to be different
        del n_lower_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys11 + 1)),
            n_lower_inner_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys22 + 1, jys22 + 1 + myg)
                ).values,
                n_lower_inner_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_lower_inner_SOL = n.bout.from_region("lower_inner_SOL")

        # Remove attributes that are expected to be different
        del n_lower_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys11 + 1)),
            n_lower_inner_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys11 + 1, jys11 + 1 + myg)
                ).values,
                n_lower_inner_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_inner_core = n.bout.from_region("inner_core")

        # Remove attributes that are expected to be different
        del n_inner_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys11 + 1, jys21 + 1)),
            n_inner_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys22 + 1 - myg, jys22 + 1)
                ).values,
                n_inner_core.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys12 + 1, jys12 + 1 + myg)
                ).values,
                n_inner_core.isel(theta=slice(-myg, None)).values,
            )

        n_inner_sol = n.bout.from_region("inner_SOL")

        # Remove attributes that are expected to be different
        del n_inner_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs - mxg, None), theta=slice(jys11 + 1, jys21 + 1)
            ),
            n_inner_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys11 + 1 - myg, jys11 + 1)
                ).values,
                n_inner_sol.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys21 + 1, jys21 + 1 + myg)
                ).values,
                n_inner_sol.isel(theta=slice(-myg, None)).values,
            )

        n_upper_inner_PFR = n.bout.from_region("upper_inner_PFR")

        # Remove attributes that are expected to be different
        del n_upper_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys21 + 1, ny_inner)),
            n_upper_inner_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys12 + 1 - myg, jys12 + 1)
                ).values,
                n_upper_inner_PFR.isel(theta=slice(myg)).values,
            )

        n_upper_inner_SOL = n.bout.from_region("upper_inner_SOL")

        # Remove attributes that are expected to be different
        del n_upper_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs - mxg, None), theta=slice(jys21 + 1, ny_inner)
            ),
            n_upper_inner_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys21 + 1 - myg, jys21 + 1)
                ).values,
                n_upper_inner_SOL.isel(theta=slice(myg)).values,
            )

        n_upper_outer_PFR = n.bout.from_region("upper_outer_PFR")

        # Remove attributes that are expected to be different
        del n_upper_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(ny_inner, jys12 + 1)),
            n_upper_outer_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys21 + 1, jys21 + 1 + myg)
                ).values,
                n_upper_outer_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_upper_outer_SOL = n.bout.from_region("upper_outer_SOL")

        # Remove attributes that are expected to be different
        del n_upper_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs - mxg, None), theta=slice(ny_inner, jys12 + 1)
            ),
            n_upper_outer_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys12 + 1, jys12 + 1 + myg)
                ).values,
                n_upper_outer_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_outer_core = n.bout.from_region("outer_core")

        # Remove attributes that are expected to be different
        del n_outer_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys12 + 1, jys22 + 1)),
            n_outer_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys21 + 1 - myg, jys21 + 1)
                ).values,
                n_outer_core.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys11 + 1, jys11 + 1 + myg)
                ).values,
                n_outer_core.isel(theta=slice(-myg, None)).values,
            )

        n_outer_sol = n.bout.from_region("outer_SOL")

        # Remove attributes that are expected to be different
        del n_outer_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs - mxg, None), theta=slice(jys12 + 1, jys22 + 1)
            ),
            n_outer_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys12 + 1 - myg, jys12 + 1)
                ).values,
                n_outer_sol.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys22 + 1, jys22 + 1 + myg)
                ).values,
                n_outer_sol.isel(theta=slice(-myg, None)).values,
            )

        n_lower_outer_PFR = n.bout.from_region("lower_outer_PFR")

        # Remove attributes that are expected to be different
        del n_lower_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs + mxg), theta=slice(jys22 + 1, None)),
            n_lower_outer_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs + mxg), theta=slice(jys11 + 1 - myg, jys11 + 1)
                ).values,
                n_lower_outer_PFR.isel(theta=slice(myg)).values,
            )

        n_lower_outer_SOL = n.bout.from_region("lower_outer_SOL")

        # Remove attributes that are expected to be different
        del n_lower_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs - mxg, None), theta=slice(jys22 + 1, None)),
            n_lower_outer_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs - mxg, None), theta=slice(jys22 + 1 - myg, jys22 + 1)
                ).values,
                n_lower_outer_SOL.isel(theta=slice(myg)).values,
            )

    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize(
        "dnd_type", ["lower-disconnected-double-null", "upper-disconnected-double-null"]
    )
    def test_region_disconnecteddoublenull(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        dnd_type,
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards=guards,
            grid="grid",
            topology=dnd_type,
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs1 = ds.metadata["ixseps1"]
        else:
            ixs1 = ds.metadata["ixseps1"] - guards["x"]

        if keep_xboundaries:
            ixs2 = ds.metadata["ixseps2"]
        else:
            ixs2 = ds.metadata["ixseps2"] - guards["x"]

        if dnd_type == "lower-disconnected-double-null":
            ixs_inner = ixs1
            ixs_outer = ixs2
        elif dnd_type == "upper-disconnected-double-null":
            ixs_inner = ixs2
            ixs_outer = ixs1
        else:
            raise ValueError(f"Unsupported dnd_type: '{dnd_type}'")

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys11 = ds.metadata["jyseps1_1"] + ybndry
        jys21 = ds.metadata["jyseps2_1"] + ybndry
        ny_inner = ds.metadata["ny_inner"] + 2 * ybndry
        jys12 = ds.metadata["jyseps1_2"] + 3 * ybndry
        jys22 = ds.metadata["jyseps2_2"] + 3 * ybndry
        ny = ds.metadata["ny"] + 4 * ybndry

        n = ds["n"]
        n_noregions = n.copy(deep=True)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_lower_inner_PFR = n.bout.from_region("lower_inner_PFR")

        # Remove attributes that are expected to be different
        del n_lower_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs_inner + mxg), theta=slice(jys11 + 1)),
            n_lower_inner_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys22 + 1, jys22 + 1 + myg)
                ).values,
                n_lower_inner_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_lower_inner_intersep = n.bout.from_region("lower_inner_intersep")

        # Remove attributes that are expected to be different
        del n_lower_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg), theta=slice(jys11 + 1)
            ),
            n_lower_inner_intersep.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys11 + 1, jys11 + 1 + myg),
                    ).values,
                    n_lower_inner_intersep.isel(theta=slice(-myg, None)).values,
                )
            else:
                # Connects to intersep in lower outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys22 + 1, jys22 + 1 + myg),
                    ).values,
                    n_lower_inner_intersep.isel(theta=slice(-myg, None)).values,
                )

        n_lower_inner_SOL = n.bout.from_region("lower_inner_SOL")

        # Remove attributes that are expected to be different
        del n_lower_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs_outer - mxg, None), theta=slice(jys11 + 1)),
            n_lower_inner_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys11 + 1, jys11 + 1 + myg),
                ).values,
                n_lower_inner_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_inner_core = n.bout.from_region("inner_core")

        # Remove attributes that are expected to be different
        del n_inner_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + mxg), theta=slice(jys11 + 1, jys21 + 1)
            ),
            n_inner_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys22 + 1 - myg, jys22 + 1)
                ).values,
                n_inner_core.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys12 + 1, jys12 + 1 + myg)
                ).values,
                n_inner_core.isel(theta=slice(-myg, None)).values,
            )

        n_inner_intersep = n.bout.from_region("inner_intersep")

        # Remove attributes that are expected to be different
        del n_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg),
                theta=slice(jys11 + 1, jys21 + 1),
            ),
            n_inner_intersep.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in lower inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys11 + 1 - myg, jys11 + 1),
                    ).values,
                    n_inner_intersep.isel(theta=slice(myg)).values,
                )

                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys12 + 1, jys12 + 1 + myg),
                    ).values,
                    n_inner_intersep.isel(theta=slice(-myg, None)).values,
                )
            else:
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys22 + 1 - myg, jys22 + 1),
                    ).values,
                    n_inner_intersep.isel(theta=slice(myg)).values,
                )

                # Connects to intersep in upper inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys21 + 1, jys21 + 1 + myg),
                    ).values,
                    n_inner_intersep.isel(theta=slice(-myg, None)).values,
                )

        n_inner_sol = n.bout.from_region("inner_SOL")

        # Remove attributes that are expected to be different
        del n_inner_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - mxg, None), theta=slice(jys11 + 1, jys21 + 1)
            ),
            n_inner_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys11 + 1 - myg, jys11 + 1),
                ).values,
                n_inner_sol.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys21 + 1, jys21 + 1 + myg),
                ).values,
                n_inner_sol.isel(theta=slice(-myg, None)).values,
            )

        n_upper_inner_PFR = n.bout.from_region("upper_inner_PFR")

        # Remove attributes that are expected to be different
        del n_upper_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + mxg), theta=slice(jys21 + 1, ny_inner)
            ),
            n_upper_inner_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys12 + 1 - myg, jys12 + 1)
                ).values,
                n_upper_inner_PFR.isel(theta=slice(myg)).values,
            )

        n_upper_inner_intersep = n.bout.from_region("upper_inner_intersep")

        # Remove attributes that are expected to be different
        del n_upper_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg),
                theta=slice(jys21 + 1, ny_inner),
            ),
            n_upper_inner_intersep.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in upper outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys12 + 1 - myg, jys12 + 1),
                    ).values,
                    n_upper_inner_intersep.isel(theta=slice(myg)).values,
                )
            else:
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys21 + 1 - myg, jys21 + 1),
                    ).values,
                    n_upper_inner_intersep.isel(theta=slice(myg)).values,
                )

        n_upper_inner_SOL = n.bout.from_region("upper_inner_SOL")

        # Remove attributes that are expected to be different
        del n_upper_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - mxg, None), theta=slice(jys21 + 1, ny_inner)
            ),
            n_upper_inner_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys21 + 1 - myg, jys21 + 1),
                ).values,
                n_upper_inner_SOL.isel(theta=slice(myg)).values,
            )

        n_upper_outer_PFR = n.bout.from_region("upper_outer_PFR")

        # Remove attributes that are expected to be different
        del n_upper_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + mxg), theta=slice(ny_inner, jys12 + 1)
            ),
            n_upper_outer_PFR.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys21 + 1, jys21 + 1 + myg)
                ).values,
                n_upper_outer_PFR.isel(theta=slice(-myg, None)).values,
            )

        n_upper_outer_intersep = n.bout.from_region("upper_outer_intersep")

        # Remove attributes that are expected to be different
        del n_upper_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg),
                theta=slice(ny_inner, jys12 + 1),
            ),
            n_upper_outer_intersep.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in upper inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys21 + 1, jys21 + 1 + myg),
                    ).values,
                    n_upper_outer_intersep.isel(theta=slice(-myg, None)).values,
                )
            else:
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys12 + 1, jys12 + 1 + myg),
                    ).values,
                    n_upper_outer_intersep.isel(theta=slice(-myg, None)).values,
                )

        n_upper_outer_SOL = n.bout.from_region("upper_outer_SOL")

        # Remove attributes that are expected to be different
        del n_upper_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - mxg, None), theta=slice(ny_inner, jys12 + 1)
            ),
            n_upper_outer_SOL.isel(theta=slice(-myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys12 + 1, jys12 + 1 + myg),
                ).values,
                n_upper_outer_SOL.isel(theta=slice(-myg, None)).values,
            )

        n_outer_core = n.bout.from_region("outer_core")

        # Remove attributes that are expected to be different
        del n_outer_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + mxg), theta=slice(jys12 + 1, jys22 + 1)
            ),
            n_outer_core.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys21 + 1 - myg, jys21 + 1)
                ).values,
                n_outer_core.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys11 + 1, jys11 + 1 + myg)
                ).values,
                n_outer_core.isel(theta=slice(-myg, None)).values,
            )

        n_outer_intersep = n.bout.from_region("outer_intersep")

        # Remove attributes that are expected to be different
        del n_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg),
                theta=slice(jys12 + 1, jys22 + 1),
            ),
            n_outer_intersep.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys21 + 1 - myg, jys21 + 1),
                    ).values,
                    n_outer_intersep.isel(theta=slice(myg)).values,
                )
                # Connects to intersep in lower outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys22 + 1, jys22 + 1 + myg),
                    ).values,
                    n_outer_intersep.isel(theta=slice(-myg, None)).values,
                )
            else:
                # Connects to intersep in upper outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys12 + 1 - myg, jys12 + 1),
                    ).values,
                    n_outer_intersep.isel(theta=slice(myg)).values,
                )
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys11 + 1, jys11 + 1 + myg),
                    ).values,
                    n_outer_intersep.isel(theta=slice(-myg, None)).values,
                )

        n_outer_sol = n.bout.from_region("outer_SOL")

        # Remove attributes that are expected to be different
        del n_outer_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - mxg, None), theta=slice(jys12 + 1, jys22 + 1)
            ),
            n_outer_sol.isel(theta=slice(myg, -myg if myg != 0 else None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys12 + 1 - myg, jys12 + 1),
                ).values,
                n_outer_sol.isel(theta=slice(myg)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys22 + 1, jys22 + 1 + myg),
                ).values,
                n_outer_sol.isel(theta=slice(-myg, None)).values,
            )

        n_lower_outer_PFR = n.bout.from_region("lower_outer_PFR")

        # Remove attributes that are expected to be different
        del n_lower_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs_inner + mxg), theta=slice(jys22 + 1, None)),
            n_lower_outer_PFR.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + mxg), theta=slice(jys11 + 1 - myg, jys11 + 1)
                ).values,
                n_lower_outer_PFR.isel(theta=slice(myg)).values,
            )

        n_lower_outer_intersep = n.bout.from_region("lower_outer_intersep")

        # Remove attributes that are expected to be different
        del n_lower_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - mxg, ixs_outer + mxg), theta=slice(jys22 + 1, None)
            ),
            n_lower_outer_intersep.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys22 + 1 - myg, jys22 + 1),
                    ).values,
                    n_lower_outer_intersep.isel(theta=slice(myg)).values,
                )
            else:
                # Connects to intersep in lower inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - mxg, ixs_outer + mxg),
                        theta=slice(jys11 + 1 - myg, jys11 + 1),
                    ).values,
                    n_lower_outer_intersep.isel(theta=slice(myg)).values,
                )

        n_lower_outer_SOL = n.bout.from_region("lower_outer_SOL")

        # Remove attributes that are expected to be different
        del n_lower_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - mxg, None), theta=slice(jys22 + 1, None)
            ),
            n_lower_outer_SOL.isel(theta=slice(myg, None)),
        )
        if myg > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - mxg, None),
                    theta=slice(jys22 + 1 - myg, jys22 + 1),
                ).values,
                n_lower_outer_SOL.isel(theta=slice(myg)).values,
            )

    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize(
        "with_guards", [0, {"x": 1}, {"theta": 1}, {"x": 1, "theta": 1}, 1]
    )
    @pytest.mark.parametrize(
        "dnd_type", ["lower-disconnected-double-null", "upper-disconnected-double-null"]
    )
    def test_region_disconnecteddoublenull_get_one_guard(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        with_guards,
        dnd_type,
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 3, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards=guards,
            grid="grid",
            topology=dnd_type,
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        mxg = guards["x"]
        myg = guards["y"]

        if keep_xboundaries:
            ixs1 = ds.metadata["ixseps1"]
        else:
            ixs1 = ds.metadata["ixseps1"] - guards["x"]

        if keep_xboundaries:
            ixs2 = ds.metadata["ixseps2"]
        else:
            ixs2 = ds.metadata["ixseps2"] - guards["x"]

        if dnd_type == "lower-disconnected-double-null":
            ixs_inner = ixs1
            ixs_outer = ixs2
        elif dnd_type == "upper-disconnected-double-null":
            ixs_inner = ixs2
            ixs_outer = ixs1
        else:
            raise ValueError(f"Unsupported dnd_type: '{dnd_type}'")

        if keep_yboundaries:
            ybndry = guards["y"]
        else:
            ybndry = 0
        jys11 = ds.metadata["jyseps1_1"] + ybndry
        jys21 = ds.metadata["jyseps2_1"] + ybndry
        ny_inner = ds.metadata["ny_inner"] + 2 * ybndry
        jys12 = ds.metadata["jyseps1_2"] + 3 * ybndry
        jys22 = ds.metadata["jyseps2_2"] + 3 * ybndry
        ny = ds.metadata["ny"] + 4 * ybndry

        if isinstance(with_guards, dict):
            xguards = with_guards.get("x", mxg)
            yguards = with_guards.get("theta", myg)
        else:
            xguards = with_guards
            yguards = with_guards

        n = ds["n"]
        n_noregions = n.copy(deep=True)

        # Remove attributes that are expected to be different
        del n_noregions.attrs["regions"]

        n_lower_inner_PFR = n.bout.from_region(
            "lower_inner_PFR", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(x=slice(ixs_inner + xguards), theta=slice(jys11 + 1)),
            n_lower_inner_PFR.isel(theta=slice(-yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys22 + 1, jys22 + 1 + yguards),
                ).values,
                n_lower_inner_PFR.isel(theta=slice(-yguards, None)).values,
            )

        n_lower_inner_intersep = n.bout.from_region(
            "lower_inner_intersep", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(jys11 + 1),
            ),
            n_lower_inner_intersep.isel(
                theta=slice(-yguards if yguards != 0 else None)
            ),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys11 + 1, jys11 + 1 + yguards),
                    ).values,
                    n_lower_inner_intersep.isel(theta=slice(-yguards, None)).values,
                )
            else:
                # Connects to intersep in lower outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys22 + 1, jys22 + 1 + yguards),
                    ).values,
                    n_lower_inner_intersep.isel(theta=slice(-yguards, None)).values,
                )

        n_lower_inner_SOL = n.bout.from_region(
            "lower_inner_SOL", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(jys11 + 1)
            ),
            n_lower_inner_SOL.isel(theta=slice(-yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys11 + 1, jys11 + 1 + yguards),
                ).values,
                n_lower_inner_SOL.isel(theta=slice(-yguards, None)).values,
            )

        n_inner_core = n.bout.from_region("inner_core", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_inner_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + xguards), theta=slice(jys11 + 1, jys21 + 1)
            ),
            n_inner_core.isel(theta=slice(yguards, -yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys22 + 1 - yguards, jys22 + 1),
                ).values,
                n_inner_core.isel(theta=slice(yguards)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys12 + 1, jys12 + 1 + yguards),
                ).values,
                n_inner_core.isel(theta=slice(-yguards, None)).values,
            )

        n_inner_intersep = n.bout.from_region("inner_intersep", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(jys11 + 1, jys21 + 1),
            ),
            n_inner_intersep.isel(
                theta=slice(yguards, -yguards if yguards != 0 else None)
            ),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in lower inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys11 + 1 - yguards, jys11 + 1),
                    ).values,
                    n_inner_intersep.isel(theta=slice(yguards)).values,
                )

                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys12 + 1, jys12 + 1 + yguards),
                    ).values,
                    n_inner_intersep.isel(theta=slice(-yguards, None)).values,
                )
            else:
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys22 + 1 - yguards, jys22 + 1),
                    ).values,
                    n_inner_intersep.isel(theta=slice(yguards)).values,
                )

                # Connects to intersep in upper inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys21 + 1, jys21 + 1 + yguards),
                    ).values,
                    n_inner_intersep.isel(theta=slice(-yguards, None)).values,
                )

        n_inner_sol = n.bout.from_region("inner_SOL", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_inner_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(jys11 + 1, jys21 + 1)
            ),
            n_inner_sol.isel(theta=slice(yguards, -yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys11 + 1 - yguards, jys11 + 1),
                ).values,
                n_inner_sol.isel(theta=slice(yguards)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys21 + 1, jys21 + 1 + yguards),
                ).values,
                n_inner_sol.isel(theta=slice(-yguards, None)).values,
            )

        n_upper_inner_PFR = n.bout.from_region(
            "upper_inner_PFR", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_inner_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + xguards), theta=slice(jys21 + 1, ny_inner)
            ),
            n_upper_inner_PFR.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys12 + 1 - yguards, jys12 + 1),
                ).values,
                n_upper_inner_PFR.isel(theta=slice(yguards)).values,
            )

        n_upper_inner_intersep = n.bout.from_region(
            "upper_inner_intersep", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_inner_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(jys21 + 1, ny_inner),
            ),
            n_upper_inner_intersep.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in upper outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys12 + 1 - yguards, jys12 + 1),
                    ).values,
                    n_upper_inner_intersep.isel(theta=slice(yguards)).values,
                )
            else:
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys21 + 1 - yguards, jys21 + 1),
                    ).values,
                    n_upper_inner_intersep.isel(theta=slice(yguards)).values,
                )

        n_upper_inner_SOL = n.bout.from_region(
            "upper_inner_SOL", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_inner_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(jys21 + 1, ny_inner)
            ),
            n_upper_inner_SOL.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys21 + 1 - yguards, jys21 + 1),
                ).values,
                n_upper_inner_SOL.isel(theta=slice(yguards)).values,
            )

        n_upper_outer_PFR = n.bout.from_region(
            "upper_outer_PFR", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + xguards), theta=slice(ny_inner, jys12 + 1)
            ),
            n_upper_outer_PFR.isel(theta=slice(-yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys21 + 1, jys21 + 1 + yguards),
                ).values,
                n_upper_outer_PFR.isel(theta=slice(-yguards, None)).values,
            )

        n_upper_outer_intersep = n.bout.from_region(
            "upper_outer_intersep", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(ny_inner, jys12 + 1),
            ),
            n_upper_outer_intersep.isel(
                theta=slice(-yguards if yguards != 0 else None)
            ),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in upper inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys21 + 1, jys21 + 1 + yguards),
                    ).values,
                    n_upper_outer_intersep.isel(theta=slice(-yguards, None)).values,
                )
            else:
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys12 + 1, jys12 + 1 + yguards),
                    ).values,
                    n_upper_outer_intersep.isel(theta=slice(-yguards, None)).values,
                )

        n_upper_outer_SOL = n.bout.from_region(
            "upper_outer_SOL", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_upper_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(ny_inner, jys12 + 1)
            ),
            n_upper_outer_SOL.isel(theta=slice(-yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys12 + 1, jys12 + 1 + yguards),
                ).values,
                n_upper_outer_SOL.isel(theta=slice(-yguards, None)).values,
            )

        n_outer_core = n.bout.from_region("outer_core", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_outer_core.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + xguards), theta=slice(jys12 + 1, jys22 + 1)
            ),
            n_outer_core.isel(theta=slice(yguards, -yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys21 + 1 - yguards, jys21 + 1),
                ).values,
                n_outer_core.isel(theta=slice(yguards)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys11 + 1, jys11 + 1 + yguards),
                ).values,
                n_outer_core.isel(theta=slice(-yguards, None)).values,
            )

        n_outer_intersep = n.bout.from_region("outer_intersep", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(jys12 + 1, jys22 + 1),
            ),
            n_outer_intersep.isel(
                theta=slice(yguards, -yguards if yguards != 0 else None)
            ),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys21 + 1 - yguards, jys21 + 1),
                    ).values,
                    n_outer_intersep.isel(theta=slice(yguards)).values,
                )

                # Connects to intersep in lower outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys22 + 1, jys22 + 1 + yguards),
                    ).values,
                    n_outer_intersep.isel(theta=slice(-yguards, None)).values,
                )
            else:
                # Connects to intersep in upper outer divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys12 + 1 - yguards, jys12 + 1),
                    ).values,
                    n_outer_intersep.isel(theta=slice(yguards)).values,
                )

                # Connects to intersep in inner SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys11 + 1, jys11 + 1 + yguards),
                    ).values,
                    n_outer_intersep.isel(theta=slice(-yguards, None)).values,
                )

        n_outer_sol = n.bout.from_region("outer_SOL", with_guards=with_guards)

        # Remove attributes that are expected to be different
        del n_outer_sol.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(jys12 + 1, jys22 + 1)
            ),
            n_outer_sol.isel(theta=slice(yguards, -yguards if yguards != 0 else None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys12 + 1 - yguards, jys12 + 1),
                ).values,
                n_outer_sol.isel(theta=slice(yguards)).values,
            )
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys22 + 1, jys22 + 1 + yguards),
                ).values,
                n_outer_sol.isel(theta=slice(-yguards, None)).values,
            )

        n_lower_outer_PFR = n.bout.from_region(
            "lower_outer_PFR", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_outer_PFR.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner + xguards), theta=slice(jys22 + 1, None)
            ),
            n_lower_outer_PFR.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_inner + xguards),
                    theta=slice(jys11 + 1 - yguards, jys11 + 1),
                ).values,
                n_lower_outer_PFR.isel(theta=slice(yguards)).values,
            )

        n_lower_outer_intersep = n.bout.from_region(
            "lower_outer_intersep", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_outer_intersep.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_inner - xguards, ixs_outer + xguards),
                theta=slice(jys22 + 1, None),
            ),
            n_lower_outer_intersep.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            if dnd_type == "lower-disconnected-double-null":
                # Connects to intersep in outer SOL
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys22 + 1 - yguards, jys22 + 1),
                    ).values,
                    n_lower_outer_intersep.isel(theta=slice(yguards)).values,
                )
            else:
                # Connects to intersep in lower inner divertor leg
                npt.assert_equal(
                    n_noregions.isel(
                        x=slice(ixs_inner - xguards, ixs_outer + xguards),
                        theta=slice(jys11 + 1 - yguards, jys11 + 1),
                    ).values,
                    n_lower_outer_intersep.isel(theta=slice(yguards)).values,
                )

        n_lower_outer_SOL = n.bout.from_region(
            "lower_outer_SOL", with_guards=with_guards
        )

        # Remove attributes that are expected to be different
        del n_lower_outer_SOL.attrs["regions"]
        xrt.assert_identical(
            n_noregions.isel(
                x=slice(ixs_outer - xguards, None), theta=slice(jys22 + 1, None)
            ),
            n_lower_outer_SOL.isel(theta=slice(yguards, None)),
        )
        if yguards > 0:
            # check y-guards, which were 'communicated' by from_region
            # Coordinates are not equal, so only compare array values
            npt.assert_equal(
                n_noregions.isel(
                    x=slice(ixs_outer - xguards, None),
                    theta=slice(jys22 + 1 - yguards, jys22 + 1),
                ).values,
                n_lower_outer_SOL.isel(theta=slice(yguards)).values,
            )



import numpy as np
from xarray import DataArray
import pytest
import numpy.testing as npt

from xbout.calc.turbulence import rms


class TestRootMeanSquare:
    def test_no_dim(self):
        dat = np.array([5, 7, 3.2, -1, -4.4])
        orig = DataArray(dat, dims=["x"])

        # Need to supply dimension
        with pytest.raises(ValueError):
            rms(orig, dim=None)

    def test_1d(self):
        dat = np.array([5, 7, 3.2, -1, -4.4])
        orig = DataArray(dat, dims=["x"])

        sum_squares = np.sum(dat**2)
        mean_squares = sum_squares / dat.size
        rootmeansquare = np.sqrt(mean_squares)

        expected = rootmeansquare
        actual = rms(orig, dim="x").values
        npt.assert_equal(actual, expected)

    @pytest.mark.parametrize("dim, axis", [("t", 1), ("x", 0)])
    def test_reduce_2d(self, dim, axis):
        dat = np.array([[5, 7, 3.2, -1, -4.4], [-1, -2.5, 0, 8, 3.0]])
        orig = DataArray(dat, dims=["x", "t"])
        sum_squares = np.sum(dat**2, axis=axis)
        mean_squares = sum_squares / dat.shape[axis]
        rootmeansquare = np.sqrt(mean_squares)

        expected = rootmeansquare
        actual = rms(orig, dim=dim).values
        npt.assert_equal(actual, expected)

    def test_reduce_2d_dask(self):
        dat = np.array([[5, 7, 3.2, -1, -4.4], [-1, -2.5, 0, 8, 3.0]])
        orig = DataArray(dat, dims=["x", "t"])
        chunked = orig.chunk({"x": 1})
        axis = 1
        sum_squares = np.sum(dat**2, axis=axis)
        mean_squares = sum_squares / dat.shape[axis]
        rootmeansquare = np.sqrt(mean_squares)

        expected = rootmeansquare
        actual = rms(chunked, dim="t").values
        npt.assert_equal(actual, expected)



from xarray import register_dataset_accessor

from xcollect.boutdataset import BoutAccessor

from wake import add_plasma_params, add_normalisations
from units import convert_units


@register_dataset_accessor("storm")
class StormAccessor(BoutAccessor):
    """
    Class specifically for holding data from a simulation using the STORM module for BOUT++.

    Implements methods for normalising the data with STORM-specific normalisation constants.
    """

    def __init__(self, ds_object):
        super().__init__(ds_object)

        # Calculate plasma parameters from options file
        # self.params = self._calc_params()

        # Calculate normalisations of data variables from parameters
        #        self._norms = self._calc_norms()

        # Set default normalisation state of data
        self.normalisation = "computational"

    def print_options(self):
        print(self.extra_data)

    # TODO as BOUT v4.2 can now write attributes to the output NetCDF files
    # would it be better to calculate all normalisations within the BOUT++ module itself?

    def _calc_params(self):
        """
        Calculates physical parameters (e.g. collisionality) from the options in the input file.

        Returns
        -------
        params : dict
        """

        params = add_plasma_params(self.options)
        return params

    def _calc_norms(self):
        """
        Calculates normalisations of physical quantities in the dataset from the options in the input file.

        Returns
        -------
        normed_ds : xarray.Dataset
        """

        normed_ds = add_normalisations(self.ds, self.options)
        return normed_ds

    def renormalise(self, desired_normalisation):
        self.data = convert_units(self.data, desired_normalisation)
        return self

    # This is where module-specific methods would go
    # For example maybe elm-pb would have a .elm_growth_rate() method?



from boutdata.data import BoutOptionsFile
from gelidum import freeze
import inspect
import numpy as np
from pathlib import Path


def load_example_input(name, *, nx=None, ny=None, nz=None):
    # Find input file in the 'inputs' subdirectory
    path = Path(__file__).parent.joinpath("inputs", name)

    return BoutOptionsFile(path, nx=nx, ny=ny, nz=nz)


def set_geometry_from_input_file(ds, name):
    ds = ds.copy()  # do not modify existing ds

    options = load_example_input(
        name, nx=ds.metadata["nx"], ny=ds.metadata["ny"], nz=ds.metadata["nz"]
    )

    # Set nx and ny in the options in case these are needed for expressions, as this is
    # not done automatically by BoutOptionsFile
    options["mesh:nx"] = ds.metadata["nx"]
    options["mesh:ny"] = ds.metadata["ny"]

    if not ds.metadata["MXG"] == options._keys.get("MXG", 2):
        raise ValueError(
            f"MXG does not match, {ds.metadata['MXG']} in ds and "
            f"{options._keys.get('MXG', 2)} in options"
        )
    if not ds.metadata["MYG"] == options._keys.get("MYG", 2):
        raise ValueError(
            f"MYG does not match, {ds.metadata['MYG']} in ds and "
            f"{options._keys.get('MYG', 2)} in options"
        )

    if ds.metadata["keep_xboundaries"] or ds.metadata["MXG"] == 0:
        slicex = slice(None)
    else:
        mxg = ds.metadata["MXG"]
        slicex = slice(mxg, -mxg)

    if ds.metadata["keep_yboundaries"] or ds.metadata["MYG"] == 0:
        slicey = slice(None)
    else:
        myg = ds.metadata["MYG"]
        slicey = slice(myg, -myg)

    slices = (slicex, slicey)

    # Replace values of fields already created by bout_xyt_example_files
    shape_2d = ds["g11"].shape
    for v in [
        "g11",
        "g22",
        "g33",
        "g12",
        "g13",
        "g23",
        "g_11",
        "g_22",
        "g_33",
        "g_12",
        "g_13",
        "g_23",
        "J",
        "Bxy",
        "zShift",
        "dx",
        "dy",
    ]:
        for location in ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]:
            suffix = "" if location == "CELL_CENTRE" else f"_{location}"
            # Need all arrays returned from options.evaluate() to be the right shape.
            # Recommend adding '0*x' or '0*y' in the input file expressions if the
            # expression would be 1d otherwise.
            ds[v + suffix] = ds[v].copy(
                data=np.broadcast_to(
                    options.evaluate(f"mesh:{v}", location=location).squeeze(axis=2)[
                        slices
                    ],
                    shape_2d,
                )
            )

    # Set dz as it would be calculated by BOUT++ (but don't support zmin, zmax or
    # zperiod here)
    if "dz" in options["mesh"]:
        ds["dz"] = options.evaluate_scalar("mesh:dz")
    else:
        ds["dz"] = 2.0 * np.pi / ds.metadata["nz"]

    # Add extra fields needed by "toroidal" geometry
    for v in ["Rxy", "Zxy", "psixy"]:
        for location in ["CELL_CENTRE", "CELL_XLOW", "CELL_YLOW", "CELL_ZLOW"]:
            suffix = "" if location == "CELL_CENTRE" else f"_{location}"
            # Need all arrays returned from options.evaluate() to be the right shape.
            # Recommend adding '0*x' or '0*y' in the input file expressions if the
            # expression would be 1d otherwise.
            ds[v + suffix] = ds["g11"].copy(
                data=np.broadcast_to(
                    options.evaluate(f"mesh:{v}", location=location).squeeze(axis=2)[
                        slices
                    ],
                    shape_2d,
                )
            )

    # Set fields that don't have to be in input files to NaN
    for v in ["G1", "G2", "G3"]:
        if v in ds:
            ds[v] = ds[v].copy(data=np.broadcast_to(float("nan"), ds[v].shape))

    return ds, options


def _get_kwargs(ignore=None):
    """
    Get the arguments of a function as a frozendict. Extended version of code from here:
    https://stackoverflow.com/a/65927265

    Parameters
    ----------
    ignore : str or Sequence of str
        Arguments to drop when constructing the returned frozendict
    """
    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    keys_to_ignore = ["self"]
    if ignore is not None:
        if isinstance(ignore, str):
            ignore = [ignore]
        keys_to_ignore += ignore
    for key in keys:
        if key not in keys_to_ignore:
            value = values[key]
            kwargs[key] = value

    return freeze(kwargs)



import pytest
import matplotlib
from matplotlib import pyplot as plt
import numpy as np
import xarray as xr

from xbout import open_boutdataset
from xbout.boutdataarray import BoutDataArrayAccessor
from .test_load import create_bout_ds_list

from animatplot.blocks import Pcolormesh, Line


@pytest.fixture
def create_test_file(tmp_path_factory):

    # Create temp dir for output of animate1D/2D
    save_dir = tmp_path_factory.mktemp("test_data")

    # Generate some test data
    ds_list, file_list = create_bout_ds_list(
        "BOUT.dmp", nxpe=3, nype=3, syn_data_type="linear"
    )
    for ds, file_name in zip(ds_list, file_list):
        ds.to_netcdf(save_dir.joinpath(file_name))

    with pytest.warns(UserWarning):
        ds = open_boutdataset(save_dir.joinpath("BOUT.dmp.*.nc"))  # Open test data

    return save_dir, ds


class TestAnimate:
    """
    Set of tests to check whether animate1D() and animate2D() are running properly
    and PillowWriter is saving each animation correctly
    """

    def test_animate2D(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds["n"].isel(x=1).bout.animate2D(save_as="%s/testyz" % save_dir)

        assert len(animation.blocks) == 1
        block = animation.blocks[0]
        assert isinstance(block, Pcolormesh)

        assert block.ax.get_xlabel() == "y"
        assert block.ax.get_ylabel() == "z"

        plt.close()

        animation = ds["n"].isel(y=2).bout.animate2D(save_as="%s/testxz" % save_dir)

        assert len(animation.blocks) == 1
        block = animation.blocks[0]
        assert isinstance(block, Pcolormesh)
        assert block.ax.get_xlabel() == "x"
        assert block.ax.get_ylabel() == "z"

        plt.close()

        animation = ds["n"].isel(z=3).bout.animate2D(save_as="%s/testxy" % save_dir)

        assert len(animation.blocks) == 1
        block = animation.blocks[0]
        assert isinstance(block, Pcolormesh)
        assert block.ax.get_xlabel() == "x"
        assert block.ax.get_ylabel() == "y"

        plt.close()

    @pytest.mark.parametrize(
        "controls",
        [
            ("both", False),
            ("timeline", False),
            ("toggle", False),
            ("", False),
            (None, False),
            ("foo", True),
        ],
    )
    def test_animate2D_controls_arg(self, create_test_file, controls):
        controls, expect_error = controls

        save_dir, ds = create_test_file

        if expect_error:
            with pytest.raises(ValueError, match="Unrecognised value for controls"):
                animation = ds["n"].isel(x=1).bout.animate2D(controls=controls)
        else:
            animation = ds["n"].isel(x=1).bout.animate2D(controls=controls)

            assert len(animation.blocks) == 1
            assert isinstance(animation.blocks[0], Pcolormesh)
            if controls in ["both", "timeline"]:
                assert hasattr(animation, "slider")
            else:
                assert not hasattr(animation, "slider")
            if controls in ["both", "toggle"]:
                assert hasattr(animation, "button")
            else:
                assert not hasattr(animation, "button")

            plt.close()

    def test_animate1D(self, create_test_file):

        save_dir, ds = create_test_file
        animation = ds["n"].isel(y=2, z=0).bout.animate1D(save_as="%s/test" % save_dir)

        assert len(animation.blocks) == 1
        assert isinstance(animation.blocks[0], Line)

        plt.close()

    @pytest.mark.parametrize(
        "controls",
        [
            ("both", False),
            ("timeline", False),
            ("toggle", False),
            ("", False),
            (None, False),
            ("foo", True),
        ],
    )
    def test_animate1D_controls_arg(self, create_test_file, controls):
        controls, expect_error = controls

        save_dir, ds = create_test_file

        if expect_error:
            with pytest.raises(ValueError, match="Unrecognised value for controls"):
                animation = ds["n"].isel(y=2, z=0).bout.animate1D(controls=controls)
        else:
            animation = ds["n"].isel(y=2, z=0).bout.animate1D(controls=controls)

            assert len(animation.blocks) == 1
            assert isinstance(animation.blocks[0], Line)
            if controls in ["both", "timeline"]:
                assert hasattr(animation, "slider")
            else:
                assert not hasattr(animation, "slider")
            if controls in ["both", "toggle"]:
                assert hasattr(animation, "button")
            else:
                assert not hasattr(animation, "button")

            plt.close()

    def test_animate_list(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)]
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_1d_default(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(y=2, z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)]
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Line)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_1d_multiline(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(y=2, z=3).bout.animate_list(
            [["n", "T"], ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)]
        )

        assert len(animation.blocks) == 4
        assert isinstance(animation.blocks[0], Line)
        assert isinstance(animation.blocks[1], Line)
        assert isinstance(animation.blocks[2], Pcolormesh)
        assert isinstance(animation.blocks[3], Line)

        # check there were actually 3 subplots
        assert (
            len(
                [
                    x
                    for x in plt.gcf().get_axes()
                    if isinstance(x, matplotlib.axes.Subplot)
                ]
            )
            == 3
        )

        plt.close()

    def test_animate_list_animate_over(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(t=2), ds["n"].isel(y=1, z=2)], animate_over="x"
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_save_as(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)],
            save_as="%s/test" % save_dir,
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_fps(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], fps=42
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)
        assert animation.timeline.fps == 42

        plt.close()

    def test_animate_list_nrows(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], nrows=2
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_ncols(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], ncols=3
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_not_enough_nrowsncols(self, create_test_file):

        save_dir, ds = create_test_file

        with pytest.raises(ValueError):
            animation = ds.isel(z=3).bout.animate_list(
                ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], nrows=2, ncols=1
            )

    @pytest.mark.skip(reason="test data for plot_poloidal needs more work")
    def test_animate_list_poloidal_plot(self, create_test_file):

        save_dir, ds = create_test_file

        metadata = ds.metadata
        metadata["ixseps1"] = 2
        metadata["ixseps2"] = 4
        metadata["jyseps1_1"] = 1
        metadata["jyseps2_1"] = 2
        metadata["ny_inner"] = 3
        metadata["jyseps1_2"] = 4
        metadata["jyseps2_2"] = 5
        from ..geometries import apply_geometry
        from ..utils import _set_attrs_on_all_vars

        ds = _set_attrs_on_all_vars(ds, "metadata", metadata)

        nx = ds.metadata["nx"]
        ny = ds.metadata["ny"]
        R = xr.DataArray(
            np.ones([nx, ny]) * np.linspace(0, 1, nx)[:, np.newaxis], dims=["x", "y"]
        )
        Z = xr.DataArray(
            np.ones([nx, ny]) * np.linspace(0, 1, ny)[np.newaxis, :], dims=["x", "y"]
        )
        ds["psixy"] = R
        ds["Rxy"] = R
        ds["Zxy"] = Z

        ds = apply_geometry(ds, "toroidal")

        animation = ds.isel(zeta=3).bout.animate_list(
            ["n", ds["T"].isel(zeta=3), ds["n"].isel(theta=1, zeta=2)],
            poloidal_plot=True,
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_subplots_adjust(self, create_test_file):

        save_dir, ds = create_test_file

        with pytest.warns(UserWarning):
            animation = ds.isel(z=3).bout.animate_list(
                ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)],
                subplots_adjust={"hspace": 4, "wspace": 5},
            )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_vmin(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], vmin=-0.1
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_vmin_list(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], vmin=[0.0, 0.1, 0.2]
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_vmax(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], vmax=1.1
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_vmax_list(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], vmax=[1.0, 1.1, 1.2]
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_logscale(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], logscale=True
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_logscale_float(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], logscale=1.0e-2
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_logscale_list(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)],
            logscale=[True, 1.0e-2, False],
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert isinstance(animation.blocks[2], Line)

        plt.close()

    def test_animate_list_titles_list(self, create_test_file):

        save_dir, ds = create_test_file

        animation = ds.isel(z=3).bout.animate_list(
            ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], titles=["a", None, "b"]
        )

        assert len(animation.blocks) == 3
        assert isinstance(animation.blocks[0], Pcolormesh)
        assert animation.blocks[0].ax.title.get_text() == "a"
        assert isinstance(animation.blocks[1], Pcolormesh)
        assert animation.blocks[1].ax.title.get_text() == "T"
        assert isinstance(animation.blocks[2], Line)
        assert animation.blocks[2].ax.title.get_text() == "b"

        plt.close()

    @pytest.mark.parametrize(
        "controls",
        [
            ("both", False),
            ("timeline", False),
            ("toggle", False),
            ("", False),
            (None, False),
            ("foo", True),
        ],
    )
    def test_animate_list_controls_arg(self, create_test_file, controls):
        controls, expect_error = controls

        save_dir, ds = create_test_file

        if expect_error:
            with pytest.raises(ValueError, match="Unrecognised value for controls"):
                animation = ds.isel(z=3).bout.animate_list(
                    ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], controls=controls
                )
        else:
            animation = ds.isel(z=3).bout.animate_list(
                ["n", ds["T"].isel(x=2), ds["n"].isel(y=1, z=2)], controls=controls
            )

            assert len(animation.blocks) == 3
            assert isinstance(animation.blocks[0], Pcolormesh)
            assert isinstance(animation.blocks[1], Pcolormesh)
            assert isinstance(animation.blocks[2], Line)
            if controls in ["both", "timeline"]:
                assert hasattr(animation, "slider")
            else:
                assert not hasattr(animation, "slider")
            if controls in ["both", "toggle"]:
                assert hasattr(animation, "button")
            else:
                assert not hasattr(animation, "button")

            plt.close()



import pytest
import numpy.testing as npt

from xbout import open_boutdataset, collect as new_collect

from .test_load import create_bout_ds, create_bout_ds_list, METADATA_VARS

boutdata = pytest.importorskip("boutdata", reason="boutdata is not available")
old_collect = boutdata.collect


class TestAccuracyAgainstOldCollect:
    def test_single_file(self, tmp_path_factory):

        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        generated_ds = create_bout_ds(syn_data_type="linear")
        generated_ds.to_netcdf(test_dir.joinpath("BOUT.dmp.0.nc"))

        var = "n"
        expected = old_collect(var, path=test_dir, xguards=True, yguards=False)

        # Test against new standard - open_boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(test_dir.joinpath("BOUT.dmp.0.nc"))
        actual = ds[var].values

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

        # Test against backwards compatible collect function
        actual = new_collect(var, path=test_dir, xguards=True, yguards=False)

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

    def test_multiple_files_along_x(self, tmp_path_factory):

        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=3, nype=1, syn_data_type="linear"
        )
        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        expected = old_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        # Test against new standard - open_boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(test_dir.joinpath("BOUT.dmp.*.nc"))
        actual = ds[var].values

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

        # Test against backwards compatible collect function
        actual = new_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

    def test_multiple_files_along_y(self, tmp_path_factory):

        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=1, nype=3, syn_data_type="linear"
        )
        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        expected = old_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        # Test against new standard - .open_boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(test_dir.joinpath("BOUT.dmp.*.nc"))
        actual = ds[var].values

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

        # Test against backwards compatible collect function
        actual = new_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

    def test_multiple_files_along_xy(self, tmp_path_factory):

        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=3, nype=3, syn_data_type="linear"
        )
        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        expected = old_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        # Test against new standard - .open_boutdataset
        with pytest.warns(UserWarning):
            ds = open_boutdataset(test_dir.joinpath("BOUT.dmp.*.nc"))
        actual = ds[var].values

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

        # Test against backwards compatible collect function
        actual = new_collect(var, path=test_dir, prefix="BOUT.dmp", xguards=True)

        assert expected.shape == actual.shape
        npt.assert_equal(actual, expected)

    def test_metadata(self, tmp_path_factory):
        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        generated_ds = create_bout_ds(syn_data_type="linear")
        generated_ds.to_netcdf(test_dir.joinpath("BOUT.dmp.0.nc"))

        with pytest.warns(UserWarning):
            ds = open_boutdataset(test_dir.joinpath("BOUT.dmp.*.nc"))

        for v in METADATA_VARS:
            expected = old_collect(v, path=test_dir)
            # Check metadata against new standard - open_boutdataset
            actual = ds.bout.metadata[v]
            npt.assert_equal(actual, expected)

            # Check against backwards compatible collect function
            actual = new_collect(v, path=test_dir)
            npt.assert_equal(actual, expected)

    def test_new_collect_indexing_int(self, tmp_path_factory):
        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=3, nype=3, syn_data_type="linear"
        )

        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        indexers = ["tind", "xind", "yind", "zind"]
        ind_arg = 0

        for kwarg in indexers:
            # Extracting a the first index of each dimension for comparison
            expected = old_collect(var, path=test_dir, **{kwarg: ind_arg})

            # Test against backwards compatible collect function
            actual = new_collect(var, path=test_dir, **{kwarg: ind_arg})

            assert expected.shape == actual.shape
            npt.assert_equal(actual, expected)

    def test_new_collect_indexing_list(self, tmp_path_factory):
        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=3, nype=3, syn_data_type="linear"
        )
        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        indexers = ["tind", "xind", "yind", "zind"]
        ind_arg = [0, 4]

        for kwarg in indexers:
            # Extracting a the first index of each dimension for comparison
            expected = old_collect(var, path=test_dir, **{kwarg: ind_arg})

            # Test against backwards compatible collect function
            actual = new_collect(var, path=test_dir, **{kwarg: ind_arg})

            assert expected.shape == actual.shape
            npt.assert_equal(actual, expected)

    def test_new_collect_indexing_slice(self, tmp_path_factory):
        # Create temp directory for files
        test_dir = tmp_path_factory.mktemp("test_data")

        # Generate some test data
        ds_list, file_list = create_bout_ds_list(
            "BOUT.dmp", nxpe=3, nype=3, syn_data_type="linear"
        )

        for temp_ds, file_name in zip(ds_list, file_list):
            temp_ds.to_netcdf(test_dir.joinpath(file_name))

        var = "n"
        indexers = ["tind", "xind", "yind", "zind"]
        ind_list = [slice(0, 4, 2), slice(0, 4)]

        for kwarg in indexers:
            for ind_arg in ind_list:
                # Extracting a the first index of each dimension for comparison
                expected = old_collect(var, path=test_dir, **{kwarg: ind_arg})

                # Test against backwards compatible collect function
                actual = new_collect(var, path=test_dir, **{kwarg: ind_arg})

                assert expected.shape == actual.shape
                npt.assert_equal(actual, expected)


@pytest.mark.skip
class test_speed_against_old_collect:
    ...



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import warnings

import xarray as xr

from .utils import (
    _create_norm,
    _decompose_regions,
    _is_core_only,
    plot_separatrices,
    plot_targets,
)


if (
    "pcolor.shading" in matplotlib.rcParams
    and matplotlib.rcParams["pcolor.shading"] == "flat"
):
    # "flat" was the old matplotlib default which discarded the last row and column if
    # X, Y and Z were all equal in size. The new "auto" should be better. Need to set
    # this explicitly because "flat" is the default during a deprecation cycle.
    matplotlib.rcParams["pcolor.shading"] = "auto"


def plot_regions(da, ax=None, **kwargs):
    """
    Plots each logical plotting region as a different color for debugging.

    Uses matplotlib.pcolormesh
    """

    x = kwargs.pop("x", "R")
    y = kwargs.pop("y", "Z")

    if len(da.dims) != 2:
        raise ValueError("da must be 2D (x,y)")

    if ax is None:
        fig, ax = plt.subplots()

    da_regions = _decompose_regions(da)

    colored_regions = [
        xr.full_like(da_region, fill_value=num / len(regions))
        for num, da_region in enumerate(da_regions.values())
    ]

    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered.
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        result = [
            region.plot.pcolormesh(
                x=x,
                y=y,
                vmin=0,
                vmax=1,
                cmap="tab20",
                infer_intervals=False,
                add_colorbar=False,
                ax=ax,
                **kwargs
            )
            for region in colored_regions
        ]

    return result


def plot2d_wrapper(
    da,
    method,
    *,
    ax=None,
    separatrix=True,
    targets=True,
    add_limiter_hatching=True,
    gridlines=None,
    cmap=None,
    norm=None,
    logscale=None,
    vmin=None,
    vmax=None,
    aspect=None,
    extend=None,
    **kwargs
):
    """
    Make a 2D plot using an xarray method, taking into account branch cuts (X-points).

    Wraps `xarray.plot` methods, so automatically adds labels.

    Parameters
    ----------
    da : xarray.DataArray
        A 2D (x,y) DataArray of data to plot
    method : xarray.plot.*
        An xarray plotting method to use
    ax : Axes, optional
        A matplotlib axes instance to plot to. If None, create a new
        figure and axes, and plot to that
    separatrix : bool, optional
        Add dashed lines showing separatrices
    targets : bool, optional
        Draw solid lines at the target surfaces
    add_limiter_hatching : bool, optional
        Draw hatched areas at the targets
    gridlines : bool, int or slice or dict of bool, int or slice, optional
        If True, draw grid lines on the plot. If an int is passed, it is used as the
        stride when plotting grid lines (to reduce the number on the plot). If a slice is
        passed it is used to select the grid lines to plot.
        If a dict is passed, the 'x' entry (bool, int or slice) is used for the radial
        grid-lines and the 'y' entry for the poloidal grid lines.
    cmap : Matplotlib colormap, optional
        Color map to use for the plot
    norm : matplotlib.colors.Normalize instance, optional
        Normalization to use for the color scale.
        Cannot be set at the same time as 'logscale'
    logscale : bool or float, optional
        If True, default to a logarithmic color scale instead of a linear one.
        If a non-bool type is passed it is treated as a float used to set the linear
        threshold of a symmetric logarithmic scale as
        linthresh=min(abs(vmin),abs(vmax))*logscale, defaults to 1e-5 if True is passed.
        Cannot be set at the same time as 'norm'
    vmin : float, optional
        Minimum value for the color scale
    vmax : float, optional
        Maximum value for the color scale
    aspect : str or float, optional
        Passed to ax.set_aspect(). By default 'equal' is used.
    extend : str or None, optional
        Passed to fig.colorbar()
    levels : int or iterable, optional
        Only used by contour or contourf, sets the number of levels (if int) or the level
        values (if iterable)
    **kwargs : optional
        Additional arguments are passed on to method

    Returns
    -------
    artists
        List of the artist instances
    """

    # TODO generalise this
    x = kwargs.pop("x", "R")
    y = kwargs.pop("y", "Z")

    if len(da.dims) != 2:
        raise ValueError("da must be 2D (x,y)")

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    if aspect is None:
        aspect = "equal"
    ax.set_aspect(aspect)

    if vmin is None:
        vmin = da.min().values
    if vmax is None:
        vmax = da.max().values

    if extend is None:
        # Replicate default for older matplotlib that does not handle extend=None
        # matplotlib-3.3 definitely does not need this. Not sure about 3.0, 3.1, 3.2.
        extend = "neither"

    # set up 'levels' if needed
    if method is xr.plot.contourf or method is xr.plot.contour:
        levels = kwargs.get("levels", 7)
        if isinstance(levels, np.int):
            levels = np.linspace(vmin, vmax, levels, endpoint=True)
            # put levels back into kwargs
            kwargs["levels"] = levels
        else:
            levels = np.array(list(levels))
            kwargs["levels"] = levels
            vmin = np.min(levels)
            vmax = np.max(levels)

        levels = kwargs.get("levels", 7)
        if isinstance(levels, np.int):
            levels = np.linspace(vmin, vmax, levels, endpoint=True)
            # put levels back into kwargs
            kwargs["levels"] = levels
        else:
            levels = np.array(list(levels))
            kwargs["levels"] = levels
            vmin = np.min(levels)
            vmax = np.max(levels)

    # Need to create a colorscale that covers the range of values in the whole array.
    # Using the add_colorbar argument would create a separate color scale for each
    # separate region, which would not make sense.
    if method is xr.plot.contourf:
        # create colorbar
        norm = _create_norm(logscale, norm, vmin, vmax)
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        # make colorbar have only discrete levels
        # average the levels so that colors in the colorbar represent the intervals
        # between the levels, as contourf colors filled regions between the given levels.
        cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            "discrete cmap",
            sm.to_rgba(0.5 * (levels[:-1] + levels[1:])),
            len(levels) - 1,
        )
        # re-make sm with new cmap
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array([])
        fig.colorbar(sm, ticks=levels, ax=ax, extend=extend)
    elif method is xr.plot.contour:
        # create colormap to be shared by all regions
        norm = matplotlib.colors.Normalize(vmin=levels[0], vmax=levels[-1])
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        cmap = matplotlib.colors.ListedColormap(
            sm.to_rgba(levels), name="discrete cmap"
        )
    else:
        # pass vmin and vmax through kwargs as they are not used for contourf or contour
        # plots
        kwargs["vmin"] = vmin
        kwargs["vmax"] = vmax

        # create colorbar
        norm = _create_norm(logscale, norm, vmin, vmax)
        sm = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
        sm.set_array([])
        cmap = sm.get_cmap()
        fig.colorbar(sm, ax=ax, extend=extend)

    if method is xr.plot.pcolormesh:
        if "infer_intervals" not in kwargs:
            kwargs["infer_intervals"] = False

    da_regions = _decompose_regions(da)

    # Plot all regions on same axis
    add_labels = [True] + [False] * (len(da_regions) - 1)
    with warnings.catch_warnings():
        # The coordinates we pass are a logically rectangular grid, so should be fine
        # even if this warning is triggered by pcolor or pcolormesh
        warnings.filterwarnings(
            "ignore",
            "The input coordinates to pcolormesh are interpreted as cell centers, but "
            "are not monotonically increasing or decreasing. This may lead to "
            "incorrectly calculated cell edges, in which case, please supply explicit "
            "cell edges to pcolormesh.",
            UserWarning,
        )
        artists = [
            method(
                region,
                x=x,
                y=y,
                ax=ax,
                add_colorbar=False,
                add_labels=add_label,
                cmap=cmap,
                **kwargs
            )
            for region, add_label in zip(da_regions.values(), add_labels)
        ]

    if method is xr.plot.contour:
        fig.colorbar(artists[0], ax=ax)

    if gridlines is not None:
        # convert gridlines to dict
        if not isinstance(gridlines, dict):
            gridlines = {"x": gridlines, "y": gridlines}

        for key, value in gridlines.items():
            if value is True:
                gridlines[key] = slice(None)
            elif isinstance(value, int):
                gridlines[key] = slice(0, None, value)
            elif value is not None:
                if not isinstance(value, slice):
                    raise ValueError(
                        "Argument passed to gridlines must be bool, int or "
                        "slice. Got a " + type(value) + ", " + str(value)
                    )

        x_regions = [da_region[x] for da_region in da_regions.values()]
        y_regions = [da_region[y] for da_region in da_regions.values()]

        for x, y in zip(x_regions, y_regions):
            if (
                not da.metadata["bout_xdim"] in x.dims
                and not da.metadata["bout_ydim"] in x.dims
            ) or (
                not da.metadata["bout_xdim"] in y.dims
                and not da.metadata["bout_ydim"] in y.dims
            ):
                # Small regions around X-point do not have segments in x- or y-directions,
                # so skip
                # Currently this region does not exist, but there is a small white gap at
                # the X-point, so we might add it back in future
                continue
            if gridlines.get("x") is not None:
                # transpose in case Dataset or DataArray has been transposed away from the usual
                # form
                dim_order = (da.metadata["bout_xdim"], da.metadata["bout_ydim"])
                yarg = {da.metadata["bout_ydim"]: gridlines["x"]}
                plt.plot(
                    x.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    y.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    color="k",
                    lw=0.1,
                )
            if gridlines.get("y") is not None:
                xarg = {da.metadata["bout_xdim"]: gridlines["y"]}
                # Need to plot transposed arrays to make gridlines that go in the
                # y-direction
                dim_order = (da.metadata["bout_ydim"], da.metadata["bout_xdim"])
                plt.plot(
                    x.isel(**xarg).transpose(*dim_order, transpose_coords=True),
                    y.isel(**yarg).transpose(*dim_order, transpose_coords=True),
                    color="k",
                    lw=0.1,
                )

    ax.set_title(da.name)

    if _is_core_only(da):
        separatrix = False
        targets = False

    if separatrix:
        plot_separatrices(da_regions, ax, x=x, y=y)

    if targets:
        plot_targets(da_regions, ax, x=x, y=y, hatching=add_limiter_hatching)

    return artists



from .load import open_boutdataset, collect

from . import geometries
from .geometries import register_geometry, REGISTERED_GEOMETRIES

from .boutdataset import BoutDatasetAccessor
from .boutdataarray import BoutDataArrayAccessor

from .plotting.animate import animate_pcolormesh, animate_poloidal
from .plotting.utils import plot_separatrix

from .fastoutput import open_fastoutput

try:
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import version, PackageNotFoundError
try:
    __version__ = version(__name__)
except PackageNotFoundError:
    from setuptools_scm import get_version

    __version__ = get_version(root="..", relative_to=__file__)



import warnings

import matplotlib as mpl
import numpy as np
import xarray as xr


def _create_norm(logscale, norm, vmin, vmax):
    if logscale:
        if norm is not None:
            raise ValueError(
                "norm and logscale cannot both be passed at the same time."
            )
        if vmin * vmax > 0:
            # vmin and vmax have the same sign, so can use standard log-scale
            norm = mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
        else:
            # vmin and vmax have opposite signs, so use symmetrical logarithmic scale
            if not isinstance(logscale, bool):
                linear_scale = logscale
            else:
                linear_scale = 1.0e-5
            linear_threshold = min(abs(vmin), abs(vmax)) * linear_scale
            norm = mpl.colors.SymLogNorm(linear_threshold, vmin=vmin, vmax=vmax)
    elif norm is None:
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    return norm


def plot_separatrix(da, sep_pos, ax, radial_coord="x"):
    """
    Plots the separatrix as a black dotted line.

    Should plot in the correct place regardless of the choice of coordinates
    for the plot axes, and the type of plot. sep_x needs to be supplied as the
    integer index of the grid point location of the separatrix.
    """

    # TODO Maybe use a decorator to do this for any type of plot?

    # 2D domain needs to intersect the separatrix plane to be able to plot it
    dims = da.dims
    if radial_coord not in dims:

        warnings.warn(
            "Cannot plot separatrix as domain does not cross "
            "separatrix, as it does not have a radial dimension",
            Warning,
        )
        return

    else:
        # Determine the separatrix position in terms of the radial coordinate
        # being used in the plot
        x_coord_vals = da.coords[radial_coord].values
        sep_x_pos = x_coord_vals[sep_pos]

        # Plot a vertical line at that location on the plot
        ax.axvline(x=sep_x_pos, linewidth=2, color="black", linestyle="--")

        return ax


def _decompose_regions(da):

    return {
        region: da.bout.from_region(region, with_guards=1)
        for region in da.bout._regions
    }


def _is_core_only(da):

    nx = da.metadata["nx"]
    ix1 = da.metadata["ixseps1"]
    ix2 = da.metadata["ixseps2"]

    return ix1 >= nx and ix2 >= nx


def plot_separatrices(da, ax, *, x="R", y="Z"):
    """Plot separatrices"""

    if not isinstance(da, dict):
        da_regions = _decompose_regions(da)
    else:
        da_regions = da

    da0 = list(da_regions.values())[0]

    xcoord = da0.metadata["bout_xdim"]
    ycoord = da0.metadata["bout_ydim"]

    for da_region in da_regions.values():
        inner = list(da_region.bout._regions.values())[0].connection_inner_x
        if inner in da_regions:
            da_inner = da_regions[inner]

            try:
                da_region, da_inner = xr.align(da_region, da_inner)
            except ValueError:
                # For geometries with a limiter, the closed field-line region may have
                # guard cells while the open field line region does not. Also the
                # closed-field line guard cells may (if the region is connected to
                # itself) have duplicated coordinate values, which xr.align() cannot
                # handle. Use np.unique() to remove the duplicated coordinate values
                _, unique_yinds = np.unique(da_inner[ycoord], return_index=True)
                da_inner = da_inner.isel(**{ycoord: unique_yinds})

            # Put da_inner second as the unique_yinds selection may mess up the order of
            # points. xarray will align the coordinates with the first argument (to the
            # addition here).
            x_sep = 0.5 * (
                da_region[x].isel(**{xcoord: 0}) + da_inner[x].isel(**{xcoord: -1})
            )
            y_sep = 0.5 * (
                da_region[y].isel(**{xcoord: 0}) + da_inner[y].isel(**{xcoord: -1})
            )
            ax.plot(x_sep, y_sep, "k--")


def plot_targets(da, ax, *, x="R", y="Z", hatching=True):
    """Plot divertor and limiter target plates"""

    if not isinstance(da, dict):
        da_regions = _decompose_regions(da)
    else:
        da_regions = da

    da0 = list(da_regions.values())[0]

    xcoord = da0.metadata["bout_xdim"]
    ycoord = da0.metadata["bout_ydim"]

    if da0.metadata["keep_yboundaries"]:
        y_boundary_guards = da0.metadata["MYG"]
    else:
        y_boundary_guards = 0

    for da_region in da_regions.values():
        if list(da_region.bout._regions.values())[0].connection_lower_y is None:
            # lower target exists
            x_target = da_region.coords[x].isel(**{ycoord: y_boundary_guards})
            y_target = da_region.coords[y].isel(**{ycoord: y_boundary_guards})
            [line] = ax.plot(x_target, y_target, "k-", linewidth=2)
            if hatching:
                _add_hatching(line, ax)
        if list(da_region.bout._regions.values())[0].connection_upper_y is None:
            # upper target exists
            x_target = da_region.coords[x].isel(**{ycoord: -y_boundary_guards - 1})
            y_target = da_region.coords[y].isel(**{ycoord: -y_boundary_guards - 1})
            [line] = ax.plot(x_target, y_target, "k-", linewidth=2)
            if hatching:
                _add_hatching(line, ax, reversed=True)


def _add_hatching(line, ax, reversed=False):
    """
    Adds a series of angled ticks to target plate lines to give a hatching
    effect, indicative of a solid surface
    """

    x = line.get_xdata()
    y = line.get_ydata()

    if reversed:
        x = np.flip(x)
        y = np.flip(y)

    # TODO redo this to evenly space ticks by physical distance along line
    num_hatchings = 3
    step = len(x) // num_hatchings
    hatch_inds = np.arange(0, len(x), step)

    vx, vy = x.max() - x.min(), y.max() - y.min()
    limiter_line_length = np.linalg.norm((vx, vy))
    hatch_line_length = limiter_line_length / num_hatchings

    # For each hatching
    for ind in hatch_inds[:-1]:
        # Compute local perpendicular vector
        dx, dy = _get_perp_vec(
            (x[ind], y[ind]), (x[ind + 1], y[ind + 1]), magnitude=hatch_line_length
        )

        # Rotate by 60 degrees
        t = -np.pi / 3
        new_dx = dx * np.cos(t) - dy * np.sin(t)
        new_dy = dx * np.sin(t) + dy * np.cos(t)

        # Draw
        ax.plot([x[ind], x[ind] + new_dx], [y[ind], y[ind] + new_dy], "k-")


def _get_perp_vec(u1, u2, magnitude=0.04):
    """Return the vector perpendicular to the vector u2-u1."""

    x1, y1 = u1
    x2, y2 = u2
    vx, vy = x2 - x1, y2 - y1
    v = np.linalg.norm((vx, vy))
    wx, wy = -vy / v * magnitude, vx / v * magnitude
    return wx, wy



import pytest

from pathlib import Path
from matplotlib import pyplot as plt
import numpy as np

from xbout import open_boutdataset
from xbout.tests.test_load import bout_xyt_example_files
from xbout.tests.test_region import (
    params_guards,
    params_guards_values,
    params_boundaries,
    params_boundaries_values,
)


class TestPlot:
    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize(
        "with_guards",
        [
            0,
            pytest.param(1, marks=pytest.mark.long),
            pytest.param(2, marks=pytest.mark.long),
        ],
    )
    @pytest.mark.parametrize(
        "dnd_type", ["lower-disconnected-double-null", "upper-disconnected-double-null"]
    )
    def test_region_disconnecteddoublenull(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        with_guards,
        dnd_type,
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 5, 4, 3),
            nxpe=3,
            nype=6,
            nt=1,
            guards=guards,
            grid="grid",
            topology=dnd_type,
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        ds["R"] = ds["R"].copy(deep=True)
        ds["R"].data[:, :] = np.linspace(0.0, 1.0, ds.sizes["x"])[:, np.newaxis]
        ds["Z"] = ds["Z"].copy(deep=True)
        ds["Z"].data[:, :] = np.linspace(0.0, 1.0, ds.sizes["theta"])[np.newaxis, :]

        n = ds["n"].isel(t=-1, zeta=0)

        n.bout.contour()
        plt.close()

        n.bout.contourf()
        plt.close()

        n.bout.pcolormesh()
        plt.close()

        n.bout.from_region("lower_inner_PFR", with_guards=with_guards).bout.pcolormesh()
        plt.close()

        n.bout.from_region(
            "lower_inner_intersep", with_guards=with_guards
        ).bout.contourf()
        plt.close()

        n.bout.from_region("lower_inner_SOL", with_guards=with_guards).bout.contour()
        plt.close()

        plt.figure()
        n.bout.from_region("inner_core", with_guards=with_guards).plot()
        plt.close()

        n.bout.from_region("inner_intersep", with_guards=with_guards).bout.pcolormesh()
        plt.close()

        n.bout.from_region("inner_SOL", with_guards=with_guards).bout.contourf()
        plt.close()

        n.bout.from_region("upper_inner_PFR", with_guards=with_guards).bout.contour()
        plt.close()

        plt.figure()
        n.bout.from_region("upper_inner_intersep", with_guards=with_guards).plot()
        plt.close()

        n.bout.from_region("upper_inner_SOL", with_guards=with_guards).bout.pcolormesh()
        plt.close()

        n.bout.from_region("upper_outer_PFR", with_guards=with_guards).bout.contourf()
        plt.close()

        n.bout.from_region(
            "upper_outer_intersep", with_guards=with_guards
        ).bout.contour()
        plt.close()

        plt.figure()
        n.bout.from_region("upper_outer_SOL", with_guards=with_guards).plot()
        plt.close()

        n.bout.from_region("outer_core", with_guards=with_guards).bout.pcolormesh()
        plt.close()

        n.bout.from_region("outer_intersep", with_guards=with_guards).bout.contourf()
        plt.close()

        n.bout.from_region("outer_SOL", with_guards=with_guards).bout.contour()
        plt.close()

        plt.figure()
        n.bout.from_region("lower_outer_PFR", with_guards=with_guards).plot()
        plt.close()

        n.bout.from_region(
            "lower_outer_intersep", with_guards=with_guards
        ).bout.pcolormesh()
        plt.close()

        n.bout.from_region("lower_outer_SOL", with_guards=with_guards).bout.contourf()
        plt.close()

    @pytest.mark.parametrize(params_guards, params_guards_values)
    @pytest.mark.parametrize(params_boundaries, params_boundaries_values)
    @pytest.mark.parametrize(
        "with_guards",
        [
            0,
            pytest.param(1, marks=pytest.mark.long),
            pytest.param(2, marks=pytest.mark.long),
        ],
    )
    def test_region_limiter(
        self,
        bout_xyt_example_files,
        guards,
        keep_xboundaries,
        keep_yboundaries,
        with_guards,
    ):
        # Note using more than MXG x-direction points and MYG y-direction points per
        # output file ensures tests for whether boundary cells are present do not fail
        # when using minimal numbers of processors
        dataset_list, grid_ds = bout_xyt_example_files(
            None,
            lengths=(2, 5, 4, 3),
            nxpe=1,
            nype=1,
            nt=1,
            guards=guards,
            grid="grid",
            topology="limiter",
        )

        ds = open_boutdataset(
            datapath=dataset_list,
            gridfilepath=grid_ds,
            geometry="toroidal",
            keep_xboundaries=keep_xboundaries,
            keep_yboundaries=keep_yboundaries,
        )

        ds["R"] = ds["R"].copy(deep=True)
        ds["R"].data[:, :] = np.linspace(0.0, 1.0, ds.sizes["x"])[:, np.newaxis]
        ds["Z"] = ds["Z"].copy(deep=True)
        ds["Z"].data[:, :] = np.linspace(0.0, 1.0, ds.sizes["theta"])[np.newaxis, :]

        n = ds["n"].isel(t=-1, zeta=0)

        n.bout.contour()
        plt.close()

        n.bout.contourf()
        plt.close()

        n.bout.pcolormesh()
        plt.close()
